@@ -14,21 +14,18 @@
 #    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 #    License for the specific language governing permissions and limitations
 #    under the License.
-"""
-OpenStack Client interface. Handles the REST calls and responses.
-"""
-
-from __future__ import print_function
+"""OpenStack Client interface. Handles the REST calls and responses."""
 
 import glob
 import hashlib
-import imp
+import importlib.util
 import itertools
 import logging
 import os
 import pkgutil
 import re
-import six
+import urllib
+from urllib import parse as urlparse
 
 from keystoneauth1 import access
 from keystoneauth1 import adapter
@@ -37,10 +34,7 @@ from keystoneauth1.identity import base
 from oslo_utils import encodeutils
 from oslo_utils import importutils
 from oslo_utils import strutils
-osprofiler_web = importutils.try_import("osprofiler.web")  # noqa
 import requests
-from six.moves import urllib
-import six.moves.urllib.parse as urlparse
 
 from cinderclient._i18n import _
 from cinderclient import api_versions
@@ -57,28 +51,40 @@ try:
 except ImportError:
     import simplejson as json
 
-_VALID_VERSIONS = ['v1', 'v2', 'v3']
+try:
+    osprofiler_web = importutils.try_import("osprofiler.web")
+except Exception:
+    pass
+
+
+_VALID_VERSIONS = ['v3']
 V3_SERVICE_TYPE = 'volumev3'
-V2_SERVICE_TYPE = 'volumev2'
-V1_SERVICE_TYPE = 'volume'
-SERVICE_TYPES = {'1': V1_SERVICE_TYPE,
-                 '2': V2_SERVICE_TYPE,
-                 '3': V3_SERVICE_TYPE}
+SERVICE_TYPES = {'3': V3_SERVICE_TYPE}
 REQ_ID_HEADER = 'X-OpenStack-Request-ID'
 
 # tell keystoneclient that we can ignore the /v1|v2/{project_id} component of
 # the service catalog when doing discovery lookups
-for svc in ('volume', 'volumev2', 'volumev3'):
+for svc in ('volume', 'volumev3'):
     discover.add_catalog_discover_hack(svc, re.compile(r'/v[12]/\w+/?$'), '/')
 
 
-def get_server_version(url):
+def get_server_version(url, insecure=False, cacert=None, cert=None):
     """Queries the server via the naked endpoint and gets version info.
 
     :param url: url of the cinder endpoint
+    :param insecure: Explicitly allow client to perform "insecure" TLS
+                     (https) requests
+    :param cacert: Specify a CA bundle file to use in verifying a TLS
+                            (https) server certificate
+    :param cert: A client certificate to pass to requests. These are of the
+                 same form as requests expects. Either a single filename
+                 containing both the certificate and key or a tuple containing
+                 the path to the certificate then a path to the key. (optional)
     :returns: APIVersion object for min and max version supported by
               the server
     """
+    # NOTE: we (the client) don't support v2 anymore, but this function
+    # is checking the server version
     min_version = "2.0"
     current_version = "2.0"
 
@@ -89,11 +95,11 @@ def get_server_version(url):
 
         # NOTE(andreykurilin): endpoint URL has at least 2 formats:
         #   1. The classic (legacy) endpoint:
-        #       http://{host}:{optional_port}/v{1 or 2 or 3}/{project-id}
-        #       http://{host}:{optional_port}/v{1 or 2 or 3}
+        #       http://{host}:{optional_port}/v{2 or 3}/{project-id}
+        #       http://{host}:{optional_port}/v{2 or 3}
         #   3. Under wsgi:
-        #       http://{host}:{optional_port}/volume/v{1 or 2 or 3}
-        for ver in ['v1', 'v2', 'v3']:
+        #       http://{host}:{optional_port}/volume/v{2 or 3}
+        for ver in ['v2', 'v3']:
             if u.path.endswith(ver) or "/{0}/".format(ver) in u.path:
                 path = u.path[:u.path.rfind(ver)]
                 version_url = '%s://%s%s' % (u.scheme, u.netloc, path)
@@ -106,7 +112,14 @@ def get_server_version(url):
             # leave as is without cropping.
             version_url = url
 
-        response = requests.get(version_url)
+        if insecure:
+            verify_cert = False
+        else:
+            if cacert:
+                verify_cert = cacert
+            else:
+                verify_cert = True
+        response = requests.get(version_url, verify=verify_cert, cert=cert)
         data = json.loads(response.text)
         versions = data['versions']
         for version in versions:
@@ -114,17 +127,38 @@ def get_server_version(url):
                 min_version = version['min_version']
                 current_version = version['version']
                 break
+            else:
+                # keep looking in case this cloud is running v2 and
+                # we haven't seen v3 yet
+                continue
     except exceptions.ClientException as e:
+        # NOTE: logging the warning but returning the lowest server API version
+        # supported in this OpenStack release is the legacy behavior, so that's
+        # what we do here
+        min_version = '3.0'
+        current_version = '3.0'
         logger.warning("Error in server version query:%s\n"
-                       "Returning APIVersion 2.0", six.text_type(e.message))
+                       "Returning APIVersion 3.0", str(e.message))
     return (api_versions.APIVersion(min_version),
             api_versions.APIVersion(current_version))
 
 
-def get_highest_client_server_version(url):
-    """Returns highest supported version by client and server as a string."""
-    min_server, max_server = get_server_version(url)
+def get_highest_client_server_version(url, insecure=False,
+                                      cacert=None, cert=None):
+    """Returns highest supported version by client and server as a string.
+
+    :raises: UnsupportedVersion if the maximum supported by the server
+             is less than the minimum supported by the client
+    """
+    min_server, max_server = get_server_version(url, insecure, cacert, cert)
     max_client = api_versions.APIVersion(api_versions.MAX_VERSION)
+    min_client = api_versions.APIVersion(api_versions.MIN_VERSION)
+    if max_server < min_client:
+        msg = _("The maximum version supported by the server (%(srv)s) does "
+                "not meet the minimum version supported by this client "
+                "(%(cli)s)") % {"srv": str(max_server),
+                                "cli": api_versions.MIN_VERSION}
+        raise exceptions.UnsupportedVersion(msg)
     return min(max_server, max_client).get_string()
 
 
@@ -141,16 +175,6 @@ def get_volume_api_from_url(url):
     raise exceptions.UnsupportedVersion(msg)
 
 
-def _log_request_id(logger, resp, service_name):
-    request_id = resp.headers.get('x-openstack-request-id')
-    if request_id:
-        logger.debug('%(method)s call to %(service_type)s for %(url)s '
-                     'used request id %(response_request_id)s',
-                     {'method': resp.request.method,
-                      'service_type': service_name,
-                      'url': resp.url, 'response_request_id': request_id})
-
-
 class SessionClient(adapter.LegacyJsonAdapter):
 
     def __init__(self, *args, **kwargs):
@@ -210,7 +234,13 @@ class SessionClient(adapter.LegacyJsonAdapter):
 
     def _get_base_url(self):
         endpoint = self.get_endpoint()
-        base_url = '/'.join(endpoint.split('/')[:3]) + '/'
+        m = re.search('(.+)/v[1-3].*', endpoint)
+        if m:
+            # Get everything up until the version identifier
+            base_url = '%s/' % m.group(1)
+        else:
+            # Fall back to the root of the URL
+            base_url = '/'.join(endpoint.split('/')[:3]) + '/'
         return base_url
 
     def get_volume_api_version_from_endpoint(self):
@@ -218,7 +248,7 @@ class SessionClient(adapter.LegacyJsonAdapter):
             version = get_volume_api_from_url(self.get_endpoint())
         except exceptions.UnsupportedVersion as e:
             msg = (_("Service catalog returned invalid url.\n"
-                     "%s") % six.text_type(e))
+                     "%s") % str(e))
             raise exceptions.UnsupportedVersion(msg)
 
         return version
@@ -248,6 +278,12 @@ class SessionClient(adapter.LegacyJsonAdapter):
     def get_with_base_url(self, url, **kwargs):
         return self._cs_request_base_url(url, 'GET', **kwargs)
 
+    def create_update_with_base_url(self, url, **kwargs):
+        return self._cs_request_base_url(url, 'PUT', **kwargs)
+
+    def delete_with_base_url(self, url, **kwargs):
+        return self._cs_request_base_url(url, 'DELETE', **kwargs)
+
 
 class HTTPClient(object):
 
@@ -259,8 +295,8 @@ class HTTPClient(object):
                  proxy_tenant_id=None, proxy_token=None, region_name=None,
                  endpoint_type='publicURL', service_type=None,
                  service_name=None, volume_service_name=None,
-                 bypass_url=None, retries=None,
-                 http_log_debug=False, cacert=None,
+                 os_endpoint=None, retries=None,
+                 http_log_debug=False, cacert=None, cert=None,
                  auth_system='keystone', auth_plugin=None, api_version=None,
                  logger=None, user_domain_name='Default',
                  project_domain_name='Default', global_request_id=None):
@@ -280,24 +316,25 @@ class HTTPClient(object):
                 raise exceptions.EndpointNotFound()
 
         self.auth_url = auth_url.rstrip('/') if auth_url else None
-        self.version = 'v1'
+        self.ks_version = 'v1'
         self.region_name = region_name
         self.endpoint_type = endpoint_type
         self.service_type = service_type
         self.service_name = service_name
         self.volume_service_name = volume_service_name
-        self.bypass_url = bypass_url.rstrip('/') if bypass_url else bypass_url
+        self.os_endpoint = os_endpoint.rstrip('/') \
+            if os_endpoint else os_endpoint
         self.retries = int(retries or 0)
         self.http_log_debug = http_log_debug
 
-        self.management_url = self.bypass_url or None
+        self.management_url = self.os_endpoint or None
         self.auth_token = None
         self.proxy_token = proxy_token
         self.proxy_tenant_id = proxy_tenant_id
         self.timeout = timeout
         self.user_domain_name = user_domain_name
         self.project_domain_name = project_domain_name
-
+        self.cert = cert
         if insecure:
             self.verify_cert = False
         else:
@@ -351,10 +388,6 @@ class HTTPClient(object):
             resp.headers,
             strutils.mask_password(resp.text))
 
-        # if service name is None then use service_type for logging
-        service = self.service_name or self.service_type
-        _log_request_id(self._logger, resp, service)
-
     def request(self, url, method, **kwargs):
         kwargs.setdefault('headers', kwargs.get('headers', {}))
         kwargs['headers']['User-Agent'] = self.USER_AGENT
@@ -378,6 +411,7 @@ class HTTPClient(object):
             method,
             url,
             verify=self.verify_cert,
+            cert=self.cert,
             **kwargs)
         self.http_log_resp(resp)
 
@@ -409,7 +443,7 @@ class HTTPClient(object):
                     url = self.management_url + url
                 resp, body = self.request(url, method, **kwargs)
                 return resp, body
-            except exceptions.BadRequest as e:
+            except exceptions.BadRequest:
                 if attempts > self.retries:
                     raise
             except exceptions.Unauthorized:
@@ -466,13 +500,12 @@ class HTTPClient(object):
         try:
             version = get_volume_api_from_url(self.management_url)
         except exceptions.UnsupportedVersion as e:
-            if self.management_url == self.bypass_url:
-                msg = (_("Invalid url was specified in --os-endpoint or "
-                         "environment variable CINDERCLIENT_BYPASS_URL.\n"
-                         "%s") % six.text_type(e))
+            if self.management_url == self.os_endpoint:
+                msg = (_("Invalid url was specified in --os-endpoint %s")
+                       % str(e))
             else:
                 msg = (_("Service catalog returned invalid url.\n"
-                         "%s") % six.text_type(e))
+                         "%s") % str(e))
 
             raise exceptions.UnsupportedVersion(msg)
 
@@ -480,6 +513,7 @@ class HTTPClient(object):
 
     def _extract_service_catalog(self, url, resp, body, extract_token=True):
         """See what the auth service told us and process the response.
+
         We may get redirected to another site, fail or actually get
         back a service catalog with a token and our endpoints.
         """
@@ -514,7 +548,7 @@ class HTTPClient(object):
                 raise
 
         elif resp.status_code == 305:
-            return resp['location']
+            return resp.headers['location']
         else:
             raise exceptions.from_response(resp, body)
 
@@ -551,7 +585,7 @@ class HTTPClient(object):
         path_parts = path.split('/')
         for part in path_parts:
             if len(part) > 0 and part[0] == 'v':
-                self.version = part
+                self.ks_version = part
                 break
 
         # TODO(sandy): Assume admin endpoint is 35357 for now.
@@ -561,7 +595,7 @@ class HTTPClient(object):
                                          path, query, frag))
 
         auth_url = self.auth_url
-        if self.version == "v2.0" or self.version == "v3":
+        if 'v2' in self.ks_version or 'v3' in self.ks_version:
             while auth_url:
                 if not self.auth_system or self.auth_system == 'keystone':
                     auth_url = self._v2_or_v3_auth(auth_url)
@@ -570,8 +604,8 @@ class HTTPClient(object):
             # existing token? If so, our actual endpoints may
             # be different than that of the admin token.
             if self.proxy_token:
-                if self.bypass_url:
-                    self.set_management_url(self.bypass_url)
+                if self.os_endpoint:
+                    self.set_management_url(self.os_endpoint)
                 else:
                     self._fetch_endpoints_from_auth(admin_url)
                 # Since keystone no longer returns the user token
@@ -590,8 +624,8 @@ class HTTPClient(object):
                     auth_url = auth_url + '/v2.0'
                 self._v2_or_v3_auth(auth_url)
 
-        if self.bypass_url:
-            self.set_management_url(self.bypass_url)
+        if self.os_endpoint:
+            self.set_management_url(self.os_endpoint)
         elif not self.management_url:
             raise exceptions.Unauthorized('Cinder Client')
 
@@ -620,7 +654,7 @@ class HTTPClient(object):
 
     def _v2_or_v3_auth(self, url):
         """Authenticate against a v2.0 auth service."""
-        if self.version == "v3":
+        if self.ks_version == "v3":
             body = {
                 "auth": {
                     "identity": {
@@ -647,11 +681,11 @@ class HTTPClient(object):
                 body['auth']['tenantName'] = self.projectid
             elif self.tenant_id:
                 body['auth']['tenantId'] = self.tenant_id
-        self._authenticate(url, body)
+        return self._authenticate(url, body)
 
     def _authenticate(self, url, body):
         """Authenticate and extract the service catalog."""
-        if self.version == 'v3':
+        if self.ks_version == 'v3':
             token_url = url + "/auth/tokens"
         else:
             token_url = url + "/tokens"
@@ -671,10 +705,10 @@ def _construct_http_client(username=None, password=None, project_id=None,
                            region_name=None, endpoint_type='publicURL',
                            service_type='volume',
                            service_name=None, volume_service_name=None,
-                           bypass_url=None, retries=None,
+                           os_endpoint=None, retries=None,
                            http_log_debug=False,
                            auth_system='keystone', auth_plugin=None,
-                           cacert=None, tenant_id=None,
+                           cacert=None, cert=None, tenant_id=None,
                            session=None,
                            auth=None, api_version=None,
                            **kwargs):
@@ -682,6 +716,8 @@ def _construct_http_client(username=None, password=None, project_id=None,
     if session:
         kwargs.setdefault('user_agent', 'python-cinderclient')
         kwargs.setdefault('interface', endpoint_type)
+        kwargs.setdefault('endpoint_override', os_endpoint)
+
         return SessionClient(session=session,
                              auth=auth,
                              service_type=service_type,
@@ -708,10 +744,11 @@ def _construct_http_client(username=None, password=None, project_id=None,
                           service_type=service_type,
                           service_name=service_name,
                           volume_service_name=volume_service_name,
-                          bypass_url=bypass_url,
+                          os_endpoint=os_endpoint,
                           retries=retries,
                           http_log_debug=http_log_debug,
                           cacert=cacert,
+                          cert=cert,
                           auth_system=auth_system,
                           auth_plugin=auth_plugin,
                           logger=logger,
@@ -733,8 +770,6 @@ def _get_client_class_and_version(version):
 
 def get_client_class(version):
     version_map = {
-        '1': 'cinderclient.v1.client.Client',
-        '2': 'cinderclient.v2.client.Client',
         '3': 'cinderclient.v3.client.Client',
     }
     try:
@@ -763,13 +798,20 @@ def _discover_via_python_path():
     for (module_loader, name, ispkg) in pkgutil.iter_modules():
         if name.endswith('cinderclient_ext'):
             if not hasattr(module_loader, 'load_module'):
-                # Python 2.6 compat: actually get an ImpImporter obj
                 module_loader = module_loader.find_module(name)
-
             module = module_loader.load_module(name)
             yield name, module
 
 
+def load_module(name, path):
+    module_spec = importlib.util.spec_from_file_location(
+        name, path
+    )
+    module = importlib.util.module_from_spec(module_spec)
+    module_spec.loader.exec_module(module)
+    return module
+
+
 def _discover_via_contrib_path(version):
     module_path = os.path.dirname(os.path.abspath(__file__))
     version_str = "v%s" % version.replace('.', '_')
@@ -782,7 +824,7 @@ def _discover_via_contrib_path(version):
         if name == "__init__":
             continue
 
-        module = imp.load_source(name, ext_path)
+        module = load_module(name, ext_path)
         yield name, module
 
 
@@ -801,8 +843,7 @@ def Client(version, *args, **kwargs):
 
     Here ``VERSION`` can be a string or
     ``cinderclient.api_versions.APIVersion`` obj. If you prefer string value,
-    you can use ``1`` (deprecated now), ``2``, or ``3.X``
-    (where X is a microversion).
+    you can use ``3`` or ``3.X`` (where X is a microversion).
 
 
     Alternatively, you can create a client instance using the keystoneclient
