@@ -39,12 +39,11 @@
 #             return original(manager, name_or_id, **kwargs)
 
 from unittest import mock
+from urllib import parse
 
 import ddt
 import fixtures
 from requests_mock.contrib import fixture as requests_mock_fixture
-import six
-from six.moves.urllib import parse
 
 import cinderclient
 from cinderclient import api_versions
@@ -52,15 +51,14 @@ from cinderclient import base
 from cinderclient import client
 from cinderclient import exceptions
 from cinderclient import shell
+from cinderclient.tests.unit.fixture_data import keystone_client
+from cinderclient.tests.unit import utils
+from cinderclient.tests.unit.v3 import fakes
 from cinderclient import utils as cinderclient_utils
 from cinderclient.v3 import attachments
 from cinderclient.v3 import volume_snapshots
 from cinderclient.v3 import volumes
 
-from cinderclient.tests.unit.fixture_data import keystone_client
-from cinderclient.tests.unit import utils
-from cinderclient.tests.unit.v3 import fakes
-
 
 @ddt.ddt
 @mock.patch.object(client, 'Client', fakes.FakeClient)
@@ -322,7 +320,7 @@ class ShellTest(utils.TestCase):
         self.assert_call_contained(
             parse.urlencode(
                 {'extra_specs':
-                    {six.text_type('key'): six.text_type('value')}}))
+                    {'key': 'value'}}))
         self.assert_call_contained(parse.urlencode({'is_public': None}))
 
     def test_type_list_public(self):
@@ -390,6 +388,25 @@ class ShellTest(utils.TestCase):
               {'cmd': 'abc 1233',
                'body': {'instance_uuid': '1233',
                         'connector': {},
+                        'volume_uuid': '1234'}},
+              {'cmd': '1234',
+               'body': {'connector': {},
+                        'volume_uuid': '1234'}},
+              {'cmd': '1234 '
+                      '--connect True '
+                      '--ip 10.23.12.23 --host server01 '
+                      '--platform x86_xx '
+                      '--ostype 123 '
+                      '--multipath true '
+                      '--mountpoint /123 '
+                      '--initiator aabbccdd',
+               'body': {'connector': {'ip': '10.23.12.23',
+                                      'host': 'server01',
+                                      'os_type': '123',
+                                      'multipath': 'true',
+                                      'mountpoint': '/123',
+                                      'initiator': 'aabbccdd',
+                                      'platform': 'x86_xx'},
                         'volume_uuid': '1234'}})
     @mock.patch('cinderclient.utils.find_resource')
     @ddt.unpack
@@ -431,6 +448,27 @@ class ShellTest(utils.TestCase):
                'body': {'instance_uuid': '1233',
                         'connector': {},
                         'volume_uuid': '1234',
+                        'mode': 'ro'}},
+              {'cmd': '1234',
+               'body': {'connector': {},
+                        'volume_uuid': '1234',
+                        'mode': 'ro'}},
+              {'cmd': '1234 '
+                      '--connect True '
+                      '--ip 10.23.12.23 --host server01 '
+                      '--platform x86_xx '
+                      '--ostype 123 '
+                      '--multipath true '
+                      '--mountpoint /123 '
+                      '--initiator aabbccdd',
+               'body': {'connector': {'ip': '10.23.12.23',
+                                      'host': 'server01',
+                                      'os_type': '123',
+                                      'multipath': 'true',
+                                      'mountpoint': '/123',
+                                      'initiator': 'aabbccdd',
+                                      'platform': 'x86_xx'},
+                        'volume_uuid': '1234',
                         'mode': 'ro'}})
     @mock.patch('cinderclient.utils.find_resource')
     @ddt.unpack
@@ -862,6 +900,67 @@ class ShellTest(utils.TestCase):
                          'manageable-list --cluster dest')
         self.assert_called('GET', '/manageable_volumes/detail?cluster=dest')
 
+    @ddt.data(True, False, 'Nonboolean')
+    @mock.patch('cinderclient.utils.find_resource')
+    def test_snapshot_create_pre_3_66(self, force_value, mock_find_vol):
+        mock_find_vol.return_value = volumes.Volume(
+            self, {'id': '123456'}, loaded=True)
+        snap_body_3_65 = {
+            'snapshot': {
+                'volume_id': '123456',
+                'force': f'{force_value}',
+                'name': None,
+                'description': None,
+                'metadata': {}
+            }
+        }
+        self.run_command('--os-volume-api-version 3.65 '
+                         f'snapshot-create --force {force_value} 123456')
+        self.assert_called_anytime('POST', '/snapshots', body=snap_body_3_65)
+
+    SNAP_BODY_3_66 = {
+        'snapshot': {
+            'volume_id': '123456',
+            'name': None,
+            'description': None,
+            'metadata': {}
+        }
+    }
+
+    @ddt.data(True, 'true', 'on', '1')
+    @mock.patch('cinderclient.utils.find_resource')
+    def test_snapshot_create_3_66_with_force_true(self, f_val, mock_find_vol):
+        mock_find_vol.return_value = volumes.Volume(
+            self, {'id': '123456'}, loaded=True)
+        mock_find_vol.return_value = volumes.Volume(self,
+                                                    {'id': '123456'},
+                                                    loaded=True)
+        self.run_command('--os-volume-api-version 3.66 '
+                         f'snapshot-create --force {f_val} 123456')
+        self.assert_called_anytime('POST', '/snapshots',
+                                   body=self.SNAP_BODY_3_66)
+
+    @ddt.data(False, 'false', 'no', '0', 'whatever')
+    @mock.patch('cinderclient.utils.find_resource')
+    def test_snapshot_create_3_66_with_force_not_true(
+            self, f_val, mock_find_vol):
+        mock_find_vol.return_value = volumes.Volume(
+            self, {'id': '123456'}, loaded=True)
+        uae = self.assertRaises(exceptions.UnsupportedAttribute,
+                                self.run_command,
+                                '--os-volume-api-version 3.66 '
+                                f'snapshot-create --force {f_val} 123456')
+        self.assertIn('not allowed after microversion 3.65', str(uae))
+
+    @mock.patch('cinderclient.utils.find_resource')
+    def test_snapshot_create_3_66(self, mock_find_vol):
+        mock_find_vol.return_value = volumes.Volume(
+            self, {'id': '123456'}, loaded=True)
+        self.run_command('--os-volume-api-version 3.66 '
+                         'snapshot-create 123456')
+        self.assert_called_anytime('POST', '/snapshots',
+                                   body=self.SNAP_BODY_3_66)
+
     def test_snapshot_manageable_list(self):
         self.run_command('--os-volume-api-version 3.8 '
                          'snapshot-manageable-list fakehost')
@@ -1150,11 +1249,7 @@ class ShellTest(utils.TestCase):
                           self.run_command, '--os-volume-api-version 3.32 '
                           'service-set-log')
         set_levels_mock.assert_not_called()
-        # Different error message from argparse library in Python 2 and 3
-        if six.PY3:
-            msg = 'the following arguments are required: <log-level>'
-        else:
-            msg = 'too few arguments'
+        msg = 'the following arguments are required: <log-level>'
         error_mock.assert_called_once_with(msg)
 
     @ddt.data('debug', 'DEBUG', 'info', 'INFO', 'warning', 'WARNING', 'error',
@@ -1284,12 +1379,14 @@ class ShellTest(utils.TestCase):
         self.assert_called('POST', '/backups', body=expected)
 
     @mock.patch("cinderclient.utils.print_list")
-    def test_snapshot_list_with_userid(self, mock_print_list):
-        """Ensure 3.41 provides User ID header."""
-        self.run_command('--os-volume-api-version 3.41 snapshot-list')
+    def test_snapshot_list(self, mock_print_list):
+        """Ensure we always present all existing fields when listing snaps."""
+        self.run_command('--os-volume-api-version 3.65 snapshot-list')
         self.assert_called('GET', '/snapshots/detail')
-        columns = ['ID', 'Volume ID', 'Status', 'Name', 'Size', 'User ID']
+        columns = ['ID', 'Volume ID', 'Status', 'Name', 'Size',
+                   'Consumes Quota', 'User ID']
         mock_print_list.assert_called_once_with(mock.ANY, columns,
+                                                exclude_unavailable=True,
                                                 sortby_index=0)
 
     @mock.patch('cinderclient.v3.volumes.Volume.migrate_volume')
@@ -1641,3 +1738,175 @@ class ShellTest(utils.TestCase):
                          '629632e7-99d2-4c40-9ae3-106fa3b1c9b7')
         self.assert_called(
             'DELETE', 'v3/default-types/629632e7-99d2-4c40-9ae3-106fa3b1c9b7')
+
+    def test_restore(self):
+        self.run_command('backup-restore 1234')
+        self.assert_called('POST', '/backups/1234/restore')
+
+    def test_restore_with_name(self):
+        self.run_command('backup-restore 1234 --name restore_vol')
+        expected = {'restore': {'volume_id': None, 'name': 'restore_vol'}}
+        self.assert_called('POST', '/backups/1234/restore',
+                           body=expected)
+
+    def test_restore_with_name_error(self):
+        self.assertRaises(exceptions.CommandError, self.run_command,
+                          'backup-restore 1234 --volume fake_vol --name '
+                          'restore_vol')
+
+    def test_restore_with_az(self):
+        self.run_command('--os-volume-api-version 3.47 backup-restore 1234 '
+                         '--name restore_vol --availability-zone restore_az')
+        expected = {'volume': {'size': 10,
+                               'name': 'restore_vol',
+                               'availability_zone': 'restore_az',
+                               'backup_id': '1234',
+                               'metadata': {},
+                               'imageRef': None,
+                               'source_volid': None,
+                               'consistencygroup_id': None,
+                               'snapshot_id': None,
+                               'volume_type': None,
+                               'description': None}}
+        self.assert_called('POST', '/volumes', body=expected)
+
+    def test_restore_with_az_microversion_error(self):
+        self.assertRaises(exceptions.UnsupportedAttribute, self.run_command,
+                          '--os-volume-api-version 3.46 backup-restore 1234 '
+                          '--name restore_vol --availability-zone restore_az')
+
+    def test_restore_with_volume_type(self):
+        self.run_command('--os-volume-api-version 3.47 backup-restore 1234 '
+                         '--name restore_vol --volume-type restore_type')
+        expected = {'volume': {'size': 10,
+                               'name': 'restore_vol',
+                               'volume_type': 'restore_type',
+                               'backup_id': '1234',
+                               'metadata': {},
+                               'imageRef': None,
+                               'source_volid': None,
+                               'consistencygroup_id': None,
+                               'snapshot_id': None,
+                               'availability_zone': None,
+                               'description': None}}
+        self.assert_called('POST', '/volumes', body=expected)
+
+    def test_restore_with_volume_type_microversion_error(self):
+        self.assertRaises(exceptions.UnsupportedAttribute, self.run_command,
+                          '--os-volume-api-version 3.46 backup-restore 1234 '
+                          '--name restore_vol --volume-type restore_type')
+
+    def test_restore_with_volume_type_and_az_no_name(self):
+        self.run_command('--os-volume-api-version 3.47 backup-restore 1234 '
+                         '--volume-type restore_type '
+                         '--availability-zone restore_az')
+        expected = {'volume': {'size': 10,
+                               'name': 'restore_backup_1234',
+                               'volume_type': 'restore_type',
+                               'availability_zone': 'restore_az',
+                               'backup_id': '1234',
+                               'metadata': {},
+                               'imageRef': None,
+                               'source_volid': None,
+                               'consistencygroup_id': None,
+                               'snapshot_id': None,
+                               'description': None}}
+        self.assert_called('POST', '/volumes', body=expected)
+
+    @ddt.data(
+        {
+            'volume': '1234',
+            'name': None,
+            'volume_type': None,
+            'availability_zone': None,
+        }, {
+            'volume': '1234',
+            'name': 'ignored',
+            'volume_type': None,
+            'availability_zone': None,
+        }, {
+            'volume': None,
+            'name': 'sample-volume',
+            'volume_type': 'sample-type',
+            'availability_zone': None,
+        }, {
+            'volume': None,
+            'name': 'sample-volume',
+            'volume_type': None,
+            'availability_zone': 'az1',
+        }, {
+            'volume': None,
+            'name': 'sample-volume',
+            'volume_type': None,
+            'availability_zone': 'different-az',
+        }, {
+            'volume': None,
+            'name': None,
+            'volume_type': None,
+            'availability_zone': 'different-az',
+        },
+    )
+    @ddt.unpack
+    @mock.patch('cinderclient.utils.print_dict')
+    @mock.patch('cinderclient.tests.unit.v3.fakes_base._stub_restore')
+    def test_do_backup_restore(self,
+                               mock_stub_restore,
+                               mock_print_dict,
+                               volume,
+                               name,
+                               volume_type,
+                               availability_zone):
+
+        # Restore from the fake '1234' backup.
+        cmd = '--os-volume-api-version 3.47 backup-restore 1234'
+
+        if volume:
+            cmd += ' --volume %s' % volume
+        if name:
+            cmd += ' --name %s' % name
+        if volume_type:
+            cmd += ' --volume-type %s' % volume_type
+        if availability_zone:
+            cmd += ' --availability-zone %s' % availability_zone
+
+        if name or volume:
+            volume_name = 'sample-volume'
+        else:
+            volume_name = 'restore_backup_1234'
+
+        mock_stub_restore.return_value = {'volume_id': '1234',
+                                          'volume_name': volume_name}
+
+        self.run_command(cmd)
+
+        # Check whether mock_stub_restore was called in order to determine
+        # whether the restore command invoked the backup-restore API. If
+        # mock_stub_restore was not called then this indicates the command
+        # invoked the volume-create API to restore the backup to a new volume
+        # of a specific volume type, or in a different AZ (the fake '1234'
+        # backup is in az1).
+        if volume_type or availability_zone == 'different-az':
+            mock_stub_restore.assert_not_called()
+        else:
+            mock_stub_restore.assert_called_once()
+
+        mock_print_dict.assert_called_once_with({
+            'backup_id': '1234',
+            'volume_id': '1234',
+            'volume_name': volume_name,
+            })
+
+    def test_reimage(self):
+        self.run_command('--os-volume-api-version 3.68 reimage 1234 1')
+        expected = {'os-reimage': {'image_id': '1',
+                                   'reimage_reserved': False}}
+        self.assert_called('POST', '/volumes/1234/action', body=expected)
+
+    @ddt.data('False', 'True')
+    def test_reimage_reserved(self, reimage_reserved):
+        self.run_command(
+            '--os-volume-api-version 3.68 reimage --reimage-reserved %s 1234 1'
+            % reimage_reserved)
+        expected = {'os-reimage': {'image_id': '1',
+                                   'reimage_reserved': reimage_reserved}}
+        self.assert_called('POST', '/volumes/1234/action', body=expected)
